;; Generated package description from age.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "age" "0.1.9.0.20250806.132339" "The Age Encryption Library" '((emacs "28.1")) :commit "e99165ef5274bc4512b8d77ba2ac208c59b5d456" :authors '(("Daiki Ueno" . "ueno@unixuser.org") ("Bas Alberts" . "bas@anti.computer")) :maintainer '("Bas Alberts" . "bas@anti.computer") :keywords '("data") :url "https://github.com/anticomputer/age.el")
